
`timescale 1ps/1ps

module tb;

// clock =======================================================================
    parameter CYC  = 2500;
    parameter CDLY = 0;

    reg clk_a;
    always #(CYC/2) clk_a = ~clk_a;
    initial begin
        clk_a = 0;
    end

    wire clk_b;
    assign #(CDLY) clk_b = clk_a;

// input/output ================================================================
    parameter BIT_WIDTH = 8;
    parameter ADR_WIDTH = 3;

    localparam ADR_DEPTH = 1 << ADR_WIDTH;
    localparam UD = 1;

    reg  [BIT_WIDTH-1:0]    din_a;
    reg  [ADR_WIDTH-1:0]    addr_a;
    reg                     en_a;
    reg                     wen_a;
    wire [BIT_WIDTH-1:0]    dout_a;
    
    reg  [BIT_WIDTH-1:0]    din_b;
    reg  [ADR_WIDTH-1:0]    addr_b;
    reg                     en_b;
    reg                     wen_b;
    wire [BIT_WIDTH-1:0]    dout_b;

// connected to dut as sinchronized signals ====================================
    reg  [BIT_WIDTH-1:0]    i_din_a;
    reg  [ADR_WIDTH-1:0]    i_addr_a;
    reg                     i_en_a;
    reg                     i_wen_a;
    
    reg  [BIT_WIDTH-1:0]    i_din_b;
    reg  [ADR_WIDTH-1:0]    i_addr_b;
    reg                     i_en_b;
    reg                     i_wen_b;

    // In real world, low through latch transfers input data when clock is low.
    // But in simulation, data or clock transition is needed to transfer first data.
    initial begin
        i_en_a=0;
        i_wen_a=0;
        i_en_b=0;
        i_wen_b=0;
    end

    // port A signals are synchronized to clk_a
    always @(posedge clk_a) begin
        i_din_a  <= din_a;
        i_addr_a <= addr_a;
        i_en_a   <= en_a;
        i_wen_a  <= wen_a;
    end

    // port B signals are synchronized to clk_b
    always @(posedge clk_b) begin
        i_din_b  <= din_b;
        i_addr_b <= addr_b;
        i_en_b   <= en_b;
        i_wen_b  <= wen_b;
    end

// tasks =======================================================================
    integer i;

    integer task_id;
    initial begin
      task_id=0;
    end

    // write from A port
    task write_from_A;
      begin
        task_id=1;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_a=1; wen_a=1; addr_a=i; din_a=i+1;
        end
        #(CYC) en_a=0; wen_a=0;
      end
    endtask

    // write from B port
    task write_from_B;
      begin
        task_id=2;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_b=1; wen_b=1; addr_b=i; din_b=i+2;
        end
        #(CYC) en_b=0; wen_b=0;
      end
    endtask

    // read from A
    task read_from_A;
      begin
        task_id=3;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_a=1; wen_a=0; addr_a=i; din_a=0;
        end
        #(CYC) en_a=0; wen_a=0;
      end
    endtask

    // read from B
    task read_from_B;
      begin
        task_id=4;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_b=1; wen_b=0; addr_b=i; din_b=0;
        end
        #(CYC) en_b=0; wen_b=0;
      end
    endtask

    // read from A and B port at different address
    task read_from_AB_diff;
      begin
        task_id=5;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_a=1; wen_a=0; addr_a=i;   din_a=0;
                 en_b=1; wen_b=0; addr_b=i+4; din_b=0;
        end
        #(CYC) en_a=0; wen_a=0;
               en_b=0; wen_b=0;
      end
    endtask

    // read from A and B port at same address
    task read_from_AB_same;
      begin
        task_id=6;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_a=1; wen_a=0; addr_a=i; din_a=0;
                 en_b=1; wen_b=0; addr_b=i; din_b=0;
        end
        #(CYC) en_a=0; wen_a=0;
               en_b=0; wen_b=0;
      end
    endtask

    // write from A and B port at different address
    task write_from_AB_diff;
      begin
        task_id=7;
        for (i=0; i<ADR_DEPTH; i=i+2) begin
          #(CYC) en_a=1; wen_a=1; addr_a=i;   din_a=i+1;
                 en_b=1; wen_b=1; addr_b=i+1; din_b=i+2;
        end
        #(CYC) en_a=0; wen_a=0;
               en_b=0; wen_b=0;
      end
    endtask

    // write from A and B port at same address
    task write_from_AB_same;
      begin
        task_id=8;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_a=1; wen_a=1; addr_a=i; din_a=i+1;
                 en_b=1; wen_b=1; addr_b=i; din_b=i+2;
        end
        #(CYC) en_a=0; wen_a=0;
               en_b=0; wen_b=0;
      end
    endtask

    // write from A and read from B port at same address
    task write_A_read_B_same;
      begin
        task_id=9;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_a=1; wen_a=1; addr_a=i; din_a=i+1;
                 en_b=1; wen_b=0; addr_b=i; din_b=0;
        end
        #(CYC) en_a=0; wen_a=0;
               en_b=0; wen_b=0;
      end
    endtask

    // read from A and write from B port at same address
    task read_A_write_B_same;
      begin
        task_id=10;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_a=1; wen_a=0; addr_a=i; din_a=0;
                 en_b=1; wen_b=1; addr_b=i; din_b=i+2;
        end
        #(CYC) en_a=0; wen_a=0;
               en_b=0; wen_b=0;
      end
    endtask

    // write from A and read from B port at different address
    task write_A_read_B_diff;
      begin
        task_id=11;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_a=1; wen_a=1; addr_a=i;   din_a=i+1;
                 en_b=1; wen_b=0; addr_b=i+4; din_b=0;
        end
        #(CYC) en_a=0; wen_a=0;
               en_b=0; wen_b=0;
      end
    endtask

    // read from A and write from B port at different address
    task read_A_write_B_diff;
      begin
        task_id=12;
        for (i=0; i<ADR_DEPTH; i=i+1) begin
          #(CYC) en_a=1; wen_a=0; addr_a=i;   din_a=0;
                 en_b=1; wen_b=1; addr_b=i+4; din_b=i+2;
        end
        #(CYC) en_a=0; wen_a=0;
               en_b=0; wen_b=0;
      end
    endtask

// +====+========+========+======++
// | ID | A port | B port | addr ||
// +====+========+========+======++
// | (1)| write  | -      | -    ||
// | (2)| -      | write  | -    ||
// +----+--------+--------+------++
// | (3)| read   | -      | -    ||
// | (4)| -      | read   | -    ||
// +----+--------+--------+------++
// | (5)| read   | read   | diff ||
// | (6)| read   | read   | same ||
// +----+--------+--------+------++
// | (7)| write  | write  | diff ||
// | (8)| write  | write  | same ||
// +----+--------+--------+------++
// | (9)| write  | read   | same ||
// |(10)| read   | write  | same ||
// +----+--------+--------+------++
// |(11)| write  | read   | diff ||
// |(12)| read   | write  | diff ||
// +====+========+========+======++

// stimulus ====================================================================
    initial begin
      #(CYC*0) en_a=0; wen_a=0; en_b=0; wen_b=0; addr_a=0;
      #(CYC*1) en_a=0; wen_a=0; en_b=0; wen_b=0; addr_a=0;

      write_from_A;        // ( 1) 1-1:write from A port
      read_from_A;         // ( 3) 1-2:read  from A port
      read_from_B;         // ( 4) 1-3:read  from B port

      write_from_B;        // ( 2) 2-1:write from B port
      read_from_A;         // ( 3) 2-2:read  from A port
      read_from_B;         // ( 4) 2-3:read  from B port

      read_from_AB_diff;   // ( 5) 3-1:read from A and B port at different address
      read_from_AB_same;   // ( 6) 3-2:read from A and B port at same address

      write_from_AB_diff;  // ( 7) 4-1:write from A and B port at different address
      read_from_AB_same;   // ( 6) 4-2:read  from A and B port at same address

      write_from_AB_same;  // ( 8) 5-1:write from A and B port at same address
      read_from_AB_same;   // ( 6) 5-2:read  from A and B port at same address (X expected)

      write_from_A;        // ( 1) 6-1:write from A port
      read_from_AB_same;   // ( 6) 6-2:read  from A and B port at same address
      write_A_read_B_same; // ( 9) 6-3:write from A and read from B port at same address (X expected)
      read_from_AB_same;   // ( 6) 6-4:read  from A and B port at same address

      read_A_write_B_same; // (10) 7-1:read from A and write from B port at same address (X expected)
      read_from_AB_same;   // ( 6) 7-2:read from A and B port at same address

      write_from_A;        // ( 1) 8-1:write from A port
      read_from_AB_same;   // ( 6) 8-2:read  from A and B port at same address
      write_A_read_B_diff; // (11) 8-3:write from A and read  from B port at different address
      read_A_write_B_diff; // (12) 8-4:read  from A and write from B port at different address
      read_from_AB_same;   // ( 6) 8-5:read  from A and B port at same address

      #(CYC) $finish();
    end


// dut =========================================================================
    com_ram_tdp_async_reg #(
        .BIT_WIDTH(BIT_WIDTH),
        .ADR_WIDTH(ADR_WIDTH)
    ) dut (
        .clk_a (clk_a),
        .din_a (i_din_a),
        .addr_a(i_addr_a),
        .en_a  (i_en_a),
        .wen_a (i_wen_a),
        .dout_a(dout_a),

        .clk_b (clk_b),
        .din_b (i_din_b),
        .addr_b(i_addr_b),
        .en_b  (i_en_b),
        .wen_b (i_wen_b),
        .dout_b(dout_b)
    );

// dump ========================================================================
    initial begin
        $dumpfile("wave.vcd");
        $dumpvars(0,tb);
    end

endmodule
